@@ -1,6 +1,7 @@
-
 /*  XMMS - Cross-platform multimedia player
- *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
+ *                           Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1999-2002  Haavard Kvaalen <havardk@sol.no>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -29,7 +30,60 @@
 #include <sys/cdio.h>
 #endif
 
+#ifdef CDROMSTOP
+# define XMMS_STOP CDROMSTOP
+#elif defined CDIOCSTOP
+# define XMMS_STOP CDIOCSTOP
+#else
+# error "No stop ioctl"
+#endif
+
+#ifdef CDIOCPAUSE
+# define XMMS_PAUSE CDIOCPAUSE
+#elif defined CDROMPAUSE
+# define XMMS_PAUSE CDROMPAUSE
+#else
+# error "No pause ioctl"
+#endif
+
+#ifdef CDIOCRESUME
+# define XMMS_RESUME CDIOCRESUME
+#elif defined CDROMRESUME
+# define XMMS_RESUME CDROMRESUME
+#else
+# error "No resume ioctl"
+#endif
+
+#ifndef CDDA_DEVICE
+# ifdef HAVE_SYS_CDIO_H
+#  ifdef __FreeBSD__
+#   define CDDA_DEVICE "/dev/acd0c"
+#  elif defined __OpenBSD__
+#   define CDDA_DEVICE "/dev/cd0c"
+#  else
+#   define CDDA_DEVICE "/vol/dev/aliases/cdrom0"
+#  endif
+# else
+#   define CDDA_DEVICE "/dev/cdrom"
+# endif
+#endif
+
+#ifndef CDDA_DIRECTORY
+# ifdef HAVE_SYS_CDIO_H
+#  ifdef __FreeBSD__
+#   define CDDA_DIRECTORY "/cdrom"
+#  elif defined __OpenBSD__
+#   define CDDA_DIRECTORY "/cdrom"
+#  else
+#   define CDDA_DIRECTORY "/cdrom/cdrom0"
+#  endif
+# else
+#   define CDDA_DIRECTORY "/mnt/cdrom"
+# endif
+#endif
+
 static char * cdda_get_title(cdda_disc_toc_t *toc, gint track);
+static gboolean stop_timeout(gpointer data);
 
 static void cdda_init(void);
 static int is_our_file(char *filename);
@@ -78,28 +132,13 @@ CDDAConfig cdda_cfg;
 static gint cdda_fd = -1;
 static gint track;
 static gboolean is_paused;
+static int pause_time;
 static cdda_disc_toc_t cd_toc;
+static int stop_timeout_id;
 
 /* Time to delay stop command in 1/10 second */
 #define STOP_DELAY 20
 
-static struct {
-	gboolean playing;
-	gboolean exit;
-	gint threads;
-} stop_info;
-
-static pthread_mutex_t stop_mutex = PTHREAD_MUTEX_INITIALIZER;
-static pthread_t stop_thread;
-
-#define STOP_LOCK() pthread_mutex_lock(&stop_mutex)
-#define STOP_UNLOCK() pthread_mutex_unlock(&stop_mutex)
-
-
-#if !defined(CDROMVOLREAD) && !(defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__)))
-static gint volume_left = 100, volume_right = 100;
-#endif
-
 InputPlugin *get_iplugin_info(void)
 {
 	cdda_ip.description = g_strdup_printf(_("CD Audio Player %s"), VERSION);
@@ -116,47 +155,27 @@ static void cdda_init(void)
 	cdda_cfg.use_oss_mixer = TRUE;
 #endif
 
-	if ((cfgfile = xmms_cfg_open_default_file()))
-	{
-		xmms_cfg_read_string(cfgfile, "CDDA", "device", &cdda_cfg.device);
-		xmms_cfg_read_string(cfgfile, "CDDA", "directory", &cdda_cfg.directory);
-		xmms_cfg_read_boolean(cfgfile, "CDDA", "use_oss_mixer", &cdda_cfg.use_oss_mixer);
-		xmms_cfg_read_boolean(cfgfile, "CDDA", "title_override", &cdda_cfg.title_override);
-		xmms_cfg_read_string(cfgfile, "CDDA", "name_format", &cdda_cfg.name_format);
-		xmms_cfg_read_boolean(cfgfile, "CDDA", "use_cddb", &cdda_cfg.use_cddb);
-		xmms_cfg_read_string(cfgfile, "CDDA", "cddb_server", &cdda_cfg.cddb_server);
-		xmms_cfg_read_int(cfgfile, "CDDA", "cddb_protocol_level", &cdda_cfg.cddb_protocol_level);
+	cfgfile = xmms_cfg_open_default_file();
+
+	xmms_cfg_read_string(cfgfile, "CDDA", "device", &cdda_cfg.device);
+	xmms_cfg_read_string(cfgfile, "CDDA", "directory", &cdda_cfg.directory);
+	xmms_cfg_read_boolean(cfgfile, "CDDA", "use_oss_mixer", &cdda_cfg.use_oss_mixer);
+	xmms_cfg_read_boolean(cfgfile, "CDDA", "title_override", &cdda_cfg.title_override);
+	xmms_cfg_read_string(cfgfile, "CDDA", "name_format", &cdda_cfg.name_format);
+	xmms_cfg_read_boolean(cfgfile, "CDDA", "use_cddb", &cdda_cfg.use_cddb);
+	xmms_cfg_read_string(cfgfile, "CDDA", "cddb_server", &cdda_cfg.cddb_server);
 #ifdef WITH_CDINDEX
-		xmms_cfg_read_boolean(cfgfile, "CDDA", "use_cdin", &cdda_cfg.use_cdin);
+	xmms_cfg_read_boolean(cfgfile, "CDDA", "use_cdin", &cdda_cfg.use_cdin);
 #else
-		cdda_cfg.use_cdin = FALSE;
+	cdda_cfg.use_cdin = FALSE;
 #endif
-		xmms_cfg_read_string(cfgfile, "CDDA", "cdin_server", &cdda_cfg.cdin_server);
-		xmms_cfg_free(cfgfile);
-	}
-#ifdef HAVE_SYS_CDIO_H
-# ifdef __FreeBSD__
+	xmms_cfg_read_string(cfgfile, "CDDA", "cdin_server", &cdda_cfg.cdin_server);
+	xmms_cfg_free(cfgfile);
+
 	if (!cdda_cfg.device)
-		cdda_cfg.device = g_strdup("/dev/acd0c");
+		cdda_cfg.device = g_strdup(CDDA_DEVICE);
 	if (!cdda_cfg.directory)
-		cdda_cfg.directory = g_strdup("/cdrom");
-# elif defined __OpenBSD__
-	if (!cdda_cfg.device)
-		cdda_cfg.device = g_strdup("/dev/cd0c");
-	if (!cdda_cfg.directory)
-		cdda_cfg.directory = g_strdup("/cdrom");
-# else
-	if (!cdda_cfg.device)
-		cdda_cfg.device = g_strdup("/vol/dev/aliases/cdrom0");
-	if (!cdda_cfg.directory)
-		cdda_cfg.directory = g_strdup("/cdrom/cdrom0");
-# endif
-#else
-	if (!cdda_cfg.device)
-		cdda_cfg.device = g_strdup("/dev/cdrom");
-	if (!cdda_cfg.directory)
-		cdda_cfg.directory = g_strdup("/mnt/cdrom/");
-#endif
+		cdda_cfg.directory = g_strdup(CDDA_DIRECTORY);
 	if (!cdda_cfg.cdin_server)
 		cdda_cfg.cdin_server = g_strdup("www.cdindex.org");
 	if (!cdda_cfg.cddb_server)
@@ -167,14 +186,12 @@ static void cdda_init(void)
 
 static void cleanup(void)
 {
-	STOP_LOCK();
-	while (stop_info.threads > 0)
+	if (stop_timeout_id)
 	{
-		STOP_UNLOCK();
-		xmms_usleep(10000);
-		STOP_LOCK();
+		gtk_timeout_remove(stop_timeout_id);
+		stop_timeout(NULL);
 	}
-	STOP_UNLOCK();
+	cddb_quit();
 }
 
 static int is_our_file(char *filename)
@@ -296,10 +313,11 @@ static void play_file(char *filename)
 			 44100 * 2 * 2 * 8, 44100, 2);
 
 	is_paused = FALSE;
-	STOP_LOCK();
-	stop_info.playing = TRUE;
-	STOP_UNLOCK();
-
+	if (stop_timeout_id)
+	{
+		gtk_timeout_remove(stop_timeout_id);
+		stop_timeout_id = 0;
+	}
 	seek(0);
 }
 
@@ -323,6 +341,11 @@ static char * cdda_get_title(cdda_disc_toc_t *toc, gint track)
 	pthread_mutex_lock(&title_mutex);
 	if (!(disc_id == cached_id && cdinfo.is_valid))
 	{
+		/*
+		 * We try to look up the disc again if the info is not
+		 * valid.  The user might have configured a new server
+		 * in the meantime.
+		 */
 		cdda_cdinfo_flush(&cdinfo);
 		cached_id = disc_id;
 
@@ -336,11 +359,11 @@ static char * cdda_get_title(cdda_disc_toc_t *toc, gint track)
 				cdda_cdinfo_write_file(disc_id, &cdinfo);
 		}
 	}
-	pthread_mutex_unlock(&title_mutex);
-	
 	XMMS_NEW_TITLEINPUT(input);
 	cdda_cdinfo_get(&cdinfo, track, &input->performer, &input->album_name,
 			&input->track_name);
+	pthread_mutex_unlock(&title_mutex);
+	
 	input->track_number = track;
 	input->file_name = input->file_path = 
 	    g_strdup_printf(_("CD Audio Track %02u"), track);
@@ -356,98 +379,349 @@ static char * cdda_get_title(cdda_disc_toc_t *toc, gint track)
 	return title;
 }
 
-static void * stop_func(void *arg)
+static gboolean stop_timeout(gpointer data)
 {
-	/*
-	 * This is run as a separate thread,
-	 * and stop the cdrom after the stop delay
-	 */
-	gint i;
-	gint fd;
-
-	STOP_LOCK();
-
-	for (i = 0; i < STOP_DELAY; i++)
-	{
-		STOP_UNLOCK();
-		xmms_usleep(100000);
-		STOP_LOCK();
-		if (stop_info.playing)
-		{
-			stop_info.threads--;
-			STOP_UNLOCK();
-			return NULL;
-		}
-		if (stop_info.exit)
-			break;
-	}
+	int fd;
 
 	fd = open(cdda_cfg.device, CDOPENFLAGS);
 	if (fd != -1)
 	{
-#ifdef CDROMSTOP
-		ioctl(fd, CDROMSTOP, 0);
-#endif
-#ifdef CDIOCSTOP
-		ioctl(fd, CDIOCSTOP, 0);
-#endif
+		ioctl(fd, XMMS_STOP, 0);
 		close(fd);
 	}
-	stop_info.threads--;
-	STOP_UNLOCK();
-	return NULL;
+	stop_timeout_id = 0;
+	return FALSE;
 }
 
 static void stop(void)
 {
-	static pthread_attr_t *attr = NULL;
-	
 	if (cdda_fd < 0)
 		return;
 
-	if (!attr)
-	{
-		attr = g_malloc(sizeof(pthread_attr_t));
-		pthread_attr_init(attr);
-		pthread_attr_setdetachstate(attr, PTHREAD_CREATE_DETACHED);
-	}
-	
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-/*  	ioctl(cdda_fd, CDIOCSTOP, 0); */
-	ioctl(cdda_fd, CDIOCPAUSE, 0);
-#else
-/*  	ioctl(cdda_fd, CDROMSTOP, 0); */
-	ioctl(cdda_fd, CDROMPAUSE);
-#endif
+	ioctl(cdda_fd, XMMS_PAUSE);
 	close(cdda_fd);
 	cdda_fd = -1;
-	STOP_LOCK();
-	stop_info.threads++;
-	stop_info.playing = FALSE;
-	STOP_UNLOCK();
-	pthread_create(&stop_thread, attr, stop_func, NULL);
+	stop_timeout_id = gtk_timeout_add(STOP_DELAY * 100, stop_timeout, NULL);
 }
 
 static void cdda_pause(short p)
 {
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-	ioctl(cdda_fd, p ? CDIOCPAUSE : CDIOCRESUME);
-#else
-	ioctl(cdda_fd, p ? CDROMPAUSE : CDROMRESUME);
-#endif
-	is_paused = (p ? TRUE : FALSE);
+	if (p)
+	{
+		pause_time = get_time();
+		ioctl(cdda_fd, XMMS_PAUSE);
+	}
+	else
+	{
+		ioctl(cdda_fd, XMMS_RESUME);
+		pause_time = -1;
+	}
+	is_paused = p;
 }
 
+#ifdef XMMS_CDROM_SOLARIS
+/*
+ * Lowlevel cdrom access, Solaris style (Solaris, Linux)
+ */
+
+static void play_ioctl(struct cdda_msf *start, struct cdda_msf *end)
+{
+	struct cdrom_msf msf;
+	
+	msf.cdmsf_min0 = start->minute;
+	msf.cdmsf_sec0 = start->second;
+	msf.cdmsf_frame0 = start->frame;
+	msf.cdmsf_min1 = end->minute;
+	msf.cdmsf_sec1 = end->second;
+	msf.cdmsf_frame1 = end->frame;
+	ioctl(cdda_fd, CDROMPLAYMSF, &msf);
+}
+
+static int get_current_frame(void)
+{
+	struct cdrom_subchnl subchnl;
+
+	subchnl.cdsc_format = CDROM_MSF;
+	ioctl(cdda_fd, CDROMSUBCHNL, &subchnl);
+
+	switch (subchnl.cdsc_audiostatus)
+	{
+		case CDROM_AUDIO_COMPLETED:
+		case CDROM_AUDIO_ERROR:
+			return -1;
+	}
+
+	return(LBA(subchnl.cdsc_absaddr.msf));
+}
+
+#if !defined(CDROMVOLREAD)
+static int volume_left = 100, volume_right = 100;
+#endif
+
+static void drive_get_volume(int *l, int *r)
+{
+#if defined(CDROMVOLREAD)
+	struct cdrom_volctrl vol;
+
+	if (cdda_fd != -1)
+	{
+		ioctl(cdda_fd, CDROMVOLREAD, &vol);
+		*l = (100 * vol.channel0) / 255;
+		*r = (100 * vol.channel1) / 255;
+	}
+#else
+	*l = volume_left;
+	*r = volume_right;
+#endif
+}
+
+static void drive_set_volume(int l, int r)
+{
+	struct cdrom_volctrl vol;
+
+	if (cdda_fd != -1)
+	{
+		vol.channel0 = vol.channel2 = (l * 255) / 100;
+		vol.channel1 = vol.channel3 = (r * 255) / 100;
+		ioctl(cdda_fd, CDROMVOLCTRL, &vol);
+	}
+#if !defined(CDROMVOLREAD)
+	volume_left = l;
+	volume_right = r;
+#endif
+}
+
+static gboolean cdda_get_toc_lowlevel(int fd, cdda_disc_toc_t *info)
+{
+	struct cdrom_tochdr tochdr;
+	struct cdrom_tocentry tocentry;
+	int i;
+
+	if (ioctl(fd, CDROMREADTOCHDR, &tochdr))
+		return FALSE;
+
+	for (i = tochdr.cdth_trk0; i <= tochdr.cdth_trk1; i++)
+	{		
+		tocentry.cdte_format = CDROM_MSF;
+		tocentry.cdte_track = i;
+		if (ioctl(fd, CDROMREADTOCENTRY, &tocentry))
+			return FALSE;
+		info->track[i].minute = tocentry.cdte_addr.msf.minute;
+		info->track[i].second = tocentry.cdte_addr.msf.second;
+		info->track[i].frame = tocentry.cdte_addr.msf.frame;
+		info->track[i].flags.data_track =
+			tocentry.cdte_ctrl == CDROM_DATA_TRACK;
+	}
+
+	/* Get the leadout track */
+	tocentry.cdte_track = CDROM_LEADOUT;
+	tocentry.cdte_format = CDROM_MSF;
+	if (ioctl(fd, CDROMREADTOCENTRY, &tocentry))
+		return FALSE;
+	info->leadout.minute = tocentry.cdte_addr.msf.minute;
+	info->leadout.second = tocentry.cdte_addr.msf.second;
+	info->leadout.frame = tocentry.cdte_addr.msf.frame;
+	
+	info->first_track = tochdr.cdth_trk0;
+	info->last_track = tochdr.cdth_trk1;
+
+	return TRUE;
+}
+
+#endif
+
+#ifdef XMMS_CDROM_BSD
+/*
+ * Lowlevel cdrom access, BSD style (FreeBSD, OpenBSD, NetBSD, Darwin)
+ */
+
+static void play_ioctl(struct cdda_msf *start, struct cdda_msf *end)
+{
+	struct ioc_play_msf msf;
+
+	msf.start_m = start->minute;
+	msf.start_s = start->second;
+	msf.start_f = start->frame;
+	msf.end_m = end->minute;
+	msf.end_s = end->second;
+	msf.end_f = end->frame;
+	ioctl(cdda_fd, CDIOCPLAYMSF, &msf);
+}
+
+static int get_current_frame(void)
+{
+	struct ioc_read_subchannel subchnl;
+	struct cd_sub_channel_info subinfo;
+	subchnl.address_format = CD_MSF_FORMAT;
+	subchnl.data_format = CD_CURRENT_POSITION;
+	subchnl.data_len = sizeof(subinfo);
+	subchnl.data = &subinfo;
+	ioctl(cdda_fd, CDIOCREADSUBCHANNEL, &subchnl);
+
+	return(LBA(subchnl.data->what.position.absaddr.msf));
+}
+
+static void drive_get_volume(int *l, int *r)
+{
+	struct ioc_vol vol;
+	
+	if (cdda_fd != -1)
+	{
+		ioctl(cdda_fd, CDIOCGETVOL, &vol);
+		*l = (100 * vol.vol[0]) / 255;
+		*r = (100 * vol.vol[1]) / 255;
+	}
+}
+
+static void drive_set_volume(int l, int r)
+{
+	struct ioc_vol vol;
+
+	if (cdda_fd != -1)
+	{
+		vol.vol[0] = vol.vol[2] = (l * 255) / 100;
+		vol.vol[1] = vol.vol[3] = (r * 255) / 100;
+		ioctl(cdda_fd, CDIOCSETVOL, &vol);
+	}
+}
+
+#ifdef XMMS_CDROM_BSD_NETBSD /* NetBSD, OpenBSD */
+
+gboolean cdda_get_toc_lowlevel(int fd, cdda_disc_toc_t *info)
+{
+	struct ioc_toc_header tochdr;
+	struct ioc_read_toc_entry tocentry;
+	int i;
+
+	if (ioctl(fd, CDIOREADTOCHEADER, &tochdr))
+		return FALSE;
+
+	for (i = tochdr.starting_track; i <= tochdr.ending_track; i++)
+	{		
+		tocentry.address_format = CD_MSF_FORMAT;
+
+		tocentry.starting_track = i;
+		if (ioctl(fd, CDIOREADTOCENTRYS, &tocentry))
+			return FALSE;
+		info->track[i].minute =
+		    tocentry.data->addr.msf.minute;
+		info->track[i].second =
+		    tocentry.data->addr.msf.second;
+		info->track[i].frame =
+		    tocentry.data->addr.msf.frame;
+		info->track[i].flags.data_track =
+		    (tocentry.data->control & 4) == 4;
+	}
+
+	/* Get the leadout track */
+	tocentry.address_format = CD_MSF_FORMAT;
+
+	tocentry.starting_track = 0xAA;
+	if (ioctl(fd, CDIOREADTOCENTRYS, &tocentry))
+		return FALSE;
+	info->leadout.minute = tocentry.data->addr.msf.minute;
+	info->leadout.second = tocentry.data->addr.msf.second;
+	info->leadout.frame = tocentry.data->addr.msf.frame;
+	
+	info->first_track = tochdr.starting_track;
+	info->last_track = tochdr.ending_track;
+
+	return TRUE;
+}
+
+#elif defined(XMMS_CDROM_BSD_DARWIN)
+
+gboolean cdda_get_toc_lowlevel(int fd, cdda_disc_toc_t *info)
+{
+	struct ioc_toc_header tochdr;
+	struct ioc_read_toc_entry tocentry;
+	int i;
+
+	if (ioctl(fd, CDIOREADTOCHEADER, &tochdr))
+		return FALSE;
+
+	for (i = tochdr.starting_track; i <= tochdr.ending_track; i++)
+	{		
+		tocentry.address_format = CD_MSF_FORMAT;
+
+		tocentry.starting_track = i;
+		if (ioctl(fd, CDIOREADTOCENTRYS, &tocentry))
+			return FALSE;
+		info->track[i].minute =
+			tocentry.data->addr[1];
+		info->track[i].second =
+			tocentry.data->addr[2];
+		info->track[i].frame =
+			tocentry.data->addr[3];
+		info->track[i].flags.data_track =
+			(tocentry.data->control & 4) == 4;
+	}
+
+	/* Get the leadout track */
+	tocentry.address_format = CD_MSF_FORMAT;
+
+	tocentry.starting_track = 0xAA;
+	if (ioctl(fd, CDIOREADTOCENTRYS, &tocentry))
+		return FALSE;
+	info->leadout.minute = tocentry.data->addr[1];
+	info->leadout.second = tocentry.data->addr[2];
+	info->leadout.frame = tocentry.data->addr[3];
+
+	return TRUE;
+}
+
+#else /* FreeBSD */
+
+gboolean cdda_get_toc_lowlevel(int fd, cdda_disc_toc_t *info)
+{
+	struct ioc_toc_header tochdr;
+	struct ioc_read_toc_single_entry tocentry;
+	int i;
+
+	if (ioctl(fd, CDIOREADTOCHEADER, &tochdr))
+		return FALSE;
+
+	for (i = tochdr.starting_track; i <= tochdr.ending_track; i++)
+	{		
+		tocentry.address_format = CD_MSF_FORMAT;
+
+		tocentry.track = i;
+		if (ioctl(fd, CDIOREADTOCENTRY, &tocentry))
+			return FALSE;
+		info->track[i].minute =
+		    tocentry.entry.addr.msf.minute;
+		info->track[i].second =
+		    tocentry.entry.addr.msf.second;
+		info->track[i].frame =
+		    tocentry.entry.addr.msf.frame;
+		info->track[i].flags.data_track =
+		    (tocentry.entry.control & 4) == 4;
+	}
+
+	/* Get the leadout track */
+	tocentry.address_format = CD_MSF_FORMAT;
+
+	tocentry.track = 0xAA;
+	if (ioctl(fd, CDIOREADTOCENTRY, &tocentry))
+		return FALSE;
+	info->leadout.minute = tocentry.entry.addr.msf.minute;
+	info->leadout.second = tocentry.entry.addr.msf.second;
+	info->leadout.frame = tocentry.entry.addr.msf.frame;
+	
+	info->first_track = tochdr.starting_track;
+	info->last_track = tochdr.ending_track;
+
+	return TRUE;
+}
+#endif
+
+#endif
+
+
 static void seek(int time)
 {
 	struct cdda_msf *end, start;
 
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-	struct ioc_play_msf msf;
-#else
-	struct cdrom_msf msf;
-#endif
-
 	start.minute = (cd_toc.track[track].minute * 60 +
 			cd_toc.track[track].second + time) / 60;
 	start.second = (cd_toc.track[track].second + time) % 60;
@@ -457,59 +731,34 @@ static void seek(int time)
 	else
 		end = &cd_toc.track[track + 1];
 
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-	msf.start_m = start.minute;
-	msf.start_s = start.second;
-	msf.start_f = start.frame;
-	msf.end_m = end->minute;
-	msf.end_s = end->second;
-	msf.end_f = end->frame;
-	ioctl(cdda_fd, CDIOCPLAYMSF, &msf);
-#else
-	msf.cdmsf_min0 = start.minute;
-	msf.cdmsf_sec0 = start.second;
-	msf.cdmsf_frame0 = start.frame;
-	msf.cdmsf_min1 = end->minute;
-	msf.cdmsf_sec1 = end->second;
-	msf.cdmsf_frame1 = end->frame;
-	ioctl(cdda_fd, CDROMPLAYMSF, &msf);
-#endif
+	play_ioctl(&start, end);
+	
 	if (is_paused)
+	{
 		cdda_pause(TRUE);
+		pause_time = time * 1000;
+	}
 }
 
 static int get_time(void)
 {
-	gint frame, start_frame, length;
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-	struct ioc_read_subchannel subchnl;
-	struct cd_sub_channel_info subinfo;
+	int frame, start_frame, length;
 
 	if (cdda_fd == -1)
 		return -1;
 
-	subchnl.address_format = CD_MSF_FORMAT;
-	subchnl.data_format = CD_CURRENT_POSITION;
-	subchnl.data_len = sizeof(subinfo);
-	subchnl.data = &subinfo;
-	ioctl(cdda_fd, CDIOCREADSUBCHANNEL, &subchnl);
+	if (is_paused && pause_time != -1)
+		return pause_time;
 
-	frame = LBA(subchnl.data->what.position.absaddr.msf);
-#else
-	struct cdrom_subchnl subchnl;
+	frame = get_current_frame();
 
-	if (cdda_fd == -1)
+	if (frame == -1)
 		return -1;
-
-	subchnl.cdsc_format = CDROM_MSF;
-	ioctl(cdda_fd, CDROMSUBCHNL, &subchnl);
-
-	frame = LBA(subchnl.cdsc_absaddr.msf);
-#endif
+	
 	start_frame = LBA(cd_toc.track[track]);
 	length = cdda_calculate_track_length(&cd_toc, track);
 
-	if (frame - start_frame >= length - 20)		/* 20 seems to work better */
+	if (frame - start_frame >= length - 20)	/* 20 seems to work better */
 		return -1;
 
 	return ((frame - start_frame) * 1000) / 75;
@@ -541,132 +790,81 @@ static void get_song_info(char *filename, char **title, int *len)
 	*title = cdda_get_title(&toc, t);
 }
 
-static void get_volume(int *l, int *r)
-{
 #if defined(HAVE_SYS_SOUNDCARD_H) || defined(HAVE_MACHINE_SOUNDCARD_H)
+static void oss_get_volume(int *l, int *r)
+{
 	int fd, devs, cmd, v;
 
-	if (cdda_cfg.use_oss_mixer)
+	fd = open(DEV_MIXER, O_RDONLY);
+	if (fd != -1)
 	{
-		fd = open(DEV_MIXER, O_RDONLY);
-		if (fd != -1)
+		ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
+		if (devs & SOUND_MASK_CD)
+			cmd = SOUND_MIXER_READ_CD;
+		else if (devs & SOUND_MASK_VOLUME)
+			cmd = SOUND_MIXER_READ_VOLUME;
+		else
 		{
-			ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
-			if (devs & SOUND_MASK_CD)
-				cmd = SOUND_MIXER_READ_CD;
-			else if (devs & SOUND_MASK_VOLUME)
-				cmd = SOUND_MIXER_READ_VOLUME;
-			else
-			{
-				close(fd);
-				return;
-			}
-			ioctl(fd, cmd, &v);
-			*r = (v & 0xFF00) >> 8;
-			*l = (v & 0x00FF);
 			close(fd);
+			return;
 		}
+		ioctl(fd, cmd, &v);
+		*r = (v & 0xFF00) >> 8;
+		*l = (v & 0x00FF);
+		close(fd);
 	}
-	else
-#endif
-	if (!cdda_cfg.use_oss_mixer)
+}
+
+static void oss_set_volume(int l, int r)
+{
+	int fd, devs, cmd, v;
+
+	fd = open(DEV_MIXER, O_RDONLY);
+	if (fd != -1)
 	{
-
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-		struct ioc_vol vol;
-
-		if (cdda_fd != -1)
+		ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
+		if (devs & SOUND_MASK_CD)
+			cmd = SOUND_MIXER_WRITE_CD;
+		else if (devs & SOUND_MASK_VOLUME)
+			cmd = SOUND_MIXER_WRITE_VOLUME;
+		else
 		{
-			ioctl(cdda_fd, CDIOCGETVOL, &vol);
-			*l = (100 * vol.vol[0]) / 255;
-			*r = (100 * vol.vol[1]) / 255;
-		}
-#elif defined(CDROMVOLREAD)
-		struct cdrom_volctrl vol;
-
-		if (cdda_fd != -1)
-		{
-			ioctl(cdda_fd, CDROMVOLREAD, &vol);
-			*l = (100 * vol.channel0) / 255;
-			*r = (100 * vol.channel1) / 255;
+			close(fd);
+			return;
 		}
+		v = (r << 8) | l;
+		ioctl(fd, cmd, &v);
+		close(fd);
+	}
+}
 #else
-		*l = volume_left;
-		*r = volume_right;
+static void oss_get_volume(int *l, int *r)
+{}
+static void oss_set_volume(int l, int r)
+{}
 #endif
 
-	}
+
+static void get_volume(int *l, int *r)
+{
+	if (cdda_cfg.use_oss_mixer)
+		oss_get_volume(l, r);
+	else
+		drive_get_volume(l, r);
 }
 
 static void set_volume(int l, int r)
 {
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-	struct ioc_vol vol;
-#else
-	struct cdrom_volctrl vol;
-#endif
-	int fd, devs, cmd, v;
-
-#if defined(HAVE_SYS_SOUNDCARD_H) || defined(HAVE_MACHINE_SOUNDCARD_H)
 	if (cdda_cfg.use_oss_mixer)
-	{
-		fd = open(DEV_MIXER, O_RDONLY);
-		if (fd != -1)
-		{
-			ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
-			if (devs & SOUND_MASK_CD)
-				cmd = SOUND_MIXER_WRITE_CD;
-			else if (devs & SOUND_MASK_VOLUME)
-				cmd = SOUND_MIXER_WRITE_VOLUME;
-			else
-			{
-				close(fd);
-				return;
-			}
-			v = (r << 8) | l;
-			ioctl(fd, cmd, &v);
-			close(fd);
-		}
-	}
+		oss_set_volume(l, r);
 	else
-#endif
-	{
-		if (cdda_fd != -1)
-		{
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-			vol.vol[0] = vol.vol[2] = (l * 255) / 100;
-			vol.vol[1] = vol.vol[3] = (r * 255) / 100;
-			ioctl(cdda_fd, CDIOCSETVOL, &vol);
-#else
-			vol.channel0 = vol.channel2 = (l * 255) / 100;
-			vol.channel1 = vol.channel3 = (r * 255) / 100;
-			ioctl(cdda_fd, CDROMVOLCTRL, &vol);
-#endif
-		}
-#if !defined(CDROMVOLREAD) && !(defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__)))
-		volume_left = l;
-		volume_right = r;
-#endif
-	}
+		drive_set_volume(l, r);
 }
 
 gboolean cdda_get_toc(cdda_disc_toc_t *info)
 {
 	gboolean retv = FALSE;
-
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-	struct ioc_toc_header tochdr;
-#if defined(__OpenBSD__)
-	struct ioc_read_toc_entry tocentry;
-#else
-	struct ioc_read_toc_single_entry tocentry;
-#endif
-#else	
-	struct cdrom_tochdr tochdr;
-	struct cdrom_tocentry tocentry;
-#endif
-	gint i;
-	gint fd;
+	int fd;
 
 	if (is_mounted(cdda_cfg.device))
 		return FALSE;
@@ -676,95 +874,8 @@ gboolean cdda_get_toc(cdda_disc_toc_t *info)
 
 	memset(info, 0, sizeof(cdda_disc_toc_t));
 
-#if defined(HAVE_SYS_CDIO_H) && (defined(__FreeBSD__) || defined(__OpenBSD__))
-	if ( ioctl(fd, CDIOREADTOCHEADER, &tochdr) )
-	    goto done;
-
-	for (i = tochdr.starting_track; i <= tochdr.ending_track; i++)
-	{		
-		tocentry.address_format = CD_MSF_FORMAT;
-
-#if defined(__OpenBSD__)
-		tocentry.starting_track = i;
-		if (ioctl(fd, CDIOREADTOCENTRYS, &tocentry))
-			goto done;
-		info->track[i].minute =
-		    tocentry.data->addr.msf.minute;
-		info->track[i].second =
-		    tocentry.data->addr.msf.second;
-		info->track[i].frame =
-		    tocentry.data->addr.msf.frame;
-		info->track[i].flags.data_track =
-		    (tocentry.data->control & 4) == 4;
-#else
-		tocentry.track = i;
-		if (ioctl(fd, CDIOREADTOCENTRY, &tocentry))
-			goto done;
-		info->track[i].minute =
-		    tocentry.entry.addr.msf.minute;
-		info->track[i].second =
-		    tocentry.entry.addr.msf.second;
-		info->track[i].frame =
-		    tocentry.entry.addr.msf.frame;
-		info->track[i].flags.data_track =
-		    (tocentry.entry.control & 4) == 4;
-#endif
-	}
-
-	/* Get the leadout track */
-	tocentry.address_format = CD_MSF_FORMAT;
-
-#if defined(__OpenBSD__)
-	tocentry.starting_track = 0xAA;
-	if (ioctl(fd, CDIOREADTOCENTRYS, &tocentry))
-		goto done;
-	info->leadout.minute = tocentry.data->addr.msf.minute;
-	info->leadout.second = tocentry.data->addr.msf.second;
-	info->leadout.frame = tocentry.data->addr.msf.frame;
-#else
-	tocentry.track = 0xAA;
-	if (ioctl(fd, CDIOREADTOCENTRY, &tocentry))
-		goto done;
-	info->leadout.minute = tocentry.entry.addr.msf.minute;
-	info->leadout.second = tocentry.entry.addr.msf.second;
-	info->leadout.frame = tocentry.entry.addr.msf.frame;
-#endif
+	retv = cdda_get_toc_lowlevel(fd, info);
 	
-	info->first_track = tochdr.starting_track;
-	info->last_track = tochdr.ending_track;
-	retv = TRUE;
-
-#else
-	if (ioctl(fd, CDROMREADTOCHDR, &tochdr))
-		goto done;
-
-	for (i = tochdr.cdth_trk0; i <= tochdr.cdth_trk1; i++)
-	{		
-		tocentry.cdte_format = CDROM_MSF;
-		tocentry.cdte_track = i;
-		if (ioctl(fd, CDROMREADTOCENTRY, &tocentry))
-			goto done;
-		info->track[i].minute = tocentry.cdte_addr.msf.minute;
-		info->track[i].second = tocentry.cdte_addr.msf.second;
-		info->track[i].frame = tocentry.cdte_addr.msf.frame;
-		info->track[i].flags.data_track = tocentry.cdte_ctrl == CDROM_DATA_TRACK;
-	}
-
-	/* Get the leadout track */
-	tocentry.cdte_track = CDROM_LEADOUT;
-	tocentry.cdte_format = CDROM_MSF;
-	if (ioctl(fd, CDROMREADTOCENTRY, &tocentry))
-		goto done;
-	info->leadout.minute = tocentry.cdte_addr.msf.minute;
-	info->leadout.second = tocentry.cdte_addr.msf.second;
-	info->leadout.frame = tocentry.cdte_addr.msf.frame;
-	
-	info->first_track = tochdr.cdth_trk0;
-	info->last_track = tochdr.cdth_trk1;
-	retv = TRUE;
-#endif
-
- done:
 	close(fd);
 
 	return retv;
