@@ -1,6 +1,6 @@
 /*********************************************************************
  * 
- *    Copyright (C) 1999,
+ *    Copyright (C) 1999, 2001, 2002,
  *    Department of Computer Science, University of Troms
  * 
  * Filename:      id3_frame_url.c
@@ -8,7 +8,7 @@
  * Author:        Espen Skoglund <espensk@stud.cs.uit.no>
  * Created at:    Tue Feb  9 21:10:45 1999
  *                
- * $Id: id3_frame_url.c,v 1.2 2000/09/24 22:33:59 havard Exp $
+ * $Id: id3_frame_url.c,v 1.6 2002/02/19 20:43:17 esk Exp $
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -25,12 +25,15 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *                
  ********************************************************************/
+#include "config.h"
+
 #include <stdio.h>
 
 #include "id3.h"
 #include "id3_header.h"
 
 
+
 /*
  * Function id3_get_url (frame)
  *
@@ -44,8 +47,7 @@ char *id3_get_url(id3_frame_t *frame)
 	return NULL;
 
     /* Check if frame is compressed */
-    if ( frame->fr_data_z && !frame->fr_data )
-	if ( id3_decompress_frame( frame ) == -1 )
+    if (id3_decompress_frame(frame) == -1)
 	    return NULL;
 
     if ( frame->fr_desc->fd_id == ID3_WXXX ) {
@@ -60,23 +62,23 @@ char *id3_get_url(id3_frame_t *frame)
 	    while ( *text != 0 )
 		text++;
 
-	    return ++text;
+	    return g_strdup(++text);
 	}
-	case ID3_ENCODING_UNICODE:
+	case ID3_ENCODING_UTF16:
 	{
-	    gint16 *text16 = (gint16 *) ((int) frame->fr_data + 1);
+	    gint16 *text16 = (gint16 *) ((glong) frame->fr_data + 1);
 
 	    while ( *text16 != 0 )
 		text16++;
 
-	    return (char *) ++text16;
+	    return g_strdup((char *) (++text16));
 	}
 	default:
 	    return NULL;
 	}
     }
 
-    return (char *) frame->fr_data;
+    return g_strdup((char *) frame->fr_data);
 }
 
 
@@ -97,9 +99,11 @@ char *id3_get_url_desc(id3_frame_t *frame)
 	return frame->fr_desc->fd_description;
 
     /* Check if frame is compressed */
-    if ( frame->fr_data_z && !frame->fr_data )
-	if ( id3_decompress_frame( frame ) == -1 )
+    if (id3_decompress_frame(frame) == -1)
 	    return NULL;
 
-    return (char *) frame->fr_data + 1;
+    if (*(guint8 *) frame->fr_data == ID3_ENCODING_ISO_8859_1)
+	return g_strdup((char *) frame->fr_data + 1);
+    else
+	return id3_utf16_to_ascii((gint16 *) ((glong) frame->fr_data + 1));
 }
