@@ -1,5 +1,7 @@
 /*  XMMS - Cross-platform multimedia player
- *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
+ *                           Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1999-2002  Haavard Kvaalen
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,12 +17,17 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
+#if defined(HAVE_CONFIG_H)
+#include "config.h"
+#endif
+
 #include <sys/stat.h>
 #include <unistd.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 #include "configfile.h"
+#include <locale.h>
 
 static ConfigSection *xmms_cfg_create_section(ConfigFile * cfg, gchar * name);
 static ConfigLine *xmms_cfg_create_string(ConfigSection * section, gchar * key, gchar * value);
@@ -186,12 +193,14 @@ gboolean xmms_cfg_read_boolean(ConfigFile * cfg, gchar * section, gchar * key, g
 
 gboolean xmms_cfg_read_float(ConfigFile * cfg, gchar * section, gchar * key, gfloat * value)
 {
-	gchar *str;
+	char *str, *locale;
 
 	if (!xmms_cfg_read_string(cfg, section, key, &str))
 		return FALSE;
 
-	*value = (gfloat) g_strtod(str, NULL);
+	locale = setlocale(LC_NUMERIC, "C");
+	*value = strtod(str, NULL);
+	setlocale(LC_NUMERIC, locale);
 	g_free(str);
 
 	return TRUE;
@@ -199,12 +208,14 @@ gboolean xmms_cfg_read_float(ConfigFile * cfg, gchar * section, gchar * key, gfl
 
 gboolean xmms_cfg_read_double(ConfigFile * cfg, gchar * section, gchar * key, gdouble * value)
 {
-	gchar *str;
+	char *str, *locale;
 
 	if (!xmms_cfg_read_string(cfg, section, key, &str))
 		return FALSE;
 
-	*value = g_strtod(str, NULL);
+	locale = setlocale(LC_NUMERIC, "C");
+	*value = strtod(str, NULL);
+	setlocale(LC_NUMERIC, locale);
 	g_free(str);
 
 	return TRUE;
@@ -246,18 +257,22 @@ void xmms_cfg_write_boolean(ConfigFile * cfg, gchar * section, gchar * key, gboo
 
 void xmms_cfg_write_float(ConfigFile * cfg, gchar * section, gchar * key, gfloat value)
 {
-	gchar *strvalue;
+	char *strvalue, *locale;
 
+	locale = setlocale(LC_NUMERIC, "C");
 	strvalue = g_strdup_printf("%g", value);
+	setlocale(LC_NUMERIC, locale);
 	xmms_cfg_write_string(cfg, section, key, strvalue);
 	g_free(strvalue);
 }
 
 void xmms_cfg_write_double(ConfigFile * cfg, gchar * section, gchar * key, gdouble value)
 {
-	gchar *strvalue;
+	char *strvalue, *locale;
 
+	locale = setlocale(LC_NUMERIC, "C");
 	strvalue = g_strdup_printf("%g", value);
+	setlocale(LC_NUMERIC, locale);
 	xmms_cfg_write_string(cfg, section, key, strvalue);
 	g_free(strvalue);
 }
@@ -306,6 +321,7 @@ void xmms_cfg_free(ConfigFile * cfg)
 		section_list = g_list_next(section_list);
 	}
 	g_list_free(cfg->sections);
+	g_free(cfg);
 }
 
 static ConfigSection *xmms_cfg_create_section(ConfigFile * cfg, gchar * name)
