@@ -1,6 +1,7 @@
 /*  XMMS - Cross-platform multimedia player
- *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
- *  Copyright (C) 1999,2000  Hvard Kvlen
+ *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
+ *                           Thomas Nilsson and 4Front Technologies
+ *  Copyright (C) 1999-2002  Haavard Kvaalen
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,14 +22,17 @@
 #ifdef HAVE_WCHAR_H
 #include <wchar.h>
 #endif
+#include <X11/Xatom.h>
 
-gint playlist_list_auto_drag_down_func(gpointer data)
+static GdkFont *playlist_list_font = NULL;
+
+static int playlist_list_auto_drag_down_func(gpointer data)
 {
 	PlayList_List *pl = data;
 
 	if (pl->pl_auto_drag_down)
 	{
-		playlist_list_move_down();
+		playlist_list_move_down(pl);
 		pl->pl_first++;
 		playlistwin_update_list();
 		return TRUE;
@@ -37,13 +41,13 @@ gint playlist_list_auto_drag_down_func(gpointer data)
 	return FALSE;
 }
 
-gint playlist_list_auto_drag_up_func(gpointer data)
+static int playlist_list_auto_drag_up_func(gpointer data)
 {
 	PlayList_List *pl = data;
 
 	if (pl->pl_auto_drag_up)
 	{
-		playlist_list_move_up();
+		playlist_list_move_up(pl);
 		pl->pl_first--;
 		playlistwin_update_list();
 		return TRUE;
@@ -52,14 +56,19 @@ gint playlist_list_auto_drag_up_func(gpointer data)
 	return FALSE;
 }
 
-void playlist_list_move_up(void)
+void playlist_list_move_up(PlayList_List *pl)
 {
 	GList *list;
 
 	PL_LOCK();
-	list = get_playlist();
+	if ((list = get_playlist()) == NULL)
+	{
+		PL_UNLOCK();
+		return;
+	}
 	if (((PlaylistEntry *) list->data)->selected)
 	{
+		/* We are at the top */
 		PL_UNLOCK();
 		return;
 	}
@@ -70,16 +79,27 @@ void playlist_list_move_up(void)
 		list = g_list_next(list);
 	}
 	PL_UNLOCK();
+	if (pl->pl_prev_selected != -1)
+		pl->pl_prev_selected--;
+	if (pl->pl_prev_min != -1)
+		pl->pl_prev_min--;
+	if (pl->pl_prev_max != -1)
+		pl->pl_prev_max--;
 }
 
-void playlist_list_move_down(void)
+void playlist_list_move_down(PlayList_List *pl)
 {
 	GList *list;
 
 	PL_LOCK();
-	list = g_list_last(get_playlist());
+	if ((list = g_list_last(get_playlist())) == NULL)
+	{
+		PL_UNLOCK();
+		return;
+	}
 	if (((PlaylistEntry *) list->data)->selected)
 	{
+		/* We are at the bottom */
 		PL_UNLOCK();
 		return;
 	}
@@ -90,89 +110,76 @@ void playlist_list_move_down(void)
 		list = g_list_previous(list);
 	}
 	PL_UNLOCK();
-}
-
-void playlist_list_select_range(gint min, gint max, gboolean sel)
-{
-	GList *list;
-	PlaylistEntry *entry;
-	int i;
-
-	if (min > max)
-	{
-		i = min;
-		min = max;
-		max = i;
-	}
-
-	list = g_list_nth(get_playlist(), min);
-	for (i = min; i <= max && list; i++)
-	{
-		entry = (PlaylistEntry *) list->data;
-		entry->selected = sel;
-		list = list->next;
-	}
+	if (pl->pl_prev_selected != -1)
+		pl->pl_prev_selected++;
+	if (pl->pl_prev_min != -1)
+		pl->pl_prev_min++;
+	if (pl->pl_prev_max != -1)
+		pl->pl_prev_max++;
 }
 
 void playlist_list_button_press_cb(GtkWidget * widget, GdkEventButton * event, PlayList_List * pl)
 {
-	GList *list;
-	PlaylistEntry *entry;
-	gint nr, y;
-
-	if (get_playlist() && inside_widget(event->x, event->y, &pl->pl_widget) && pl->pl_fheight)
+	if (event->button == 1 && pl->pl_fheight &&
+	    inside_widget(event->x, event->y, &pl->pl_widget))
 	{
-		if (event->button == 1)
-		{
-			PL_LOCK();
-			y = event->y - pl->pl_widget.y;
-			nr = (y / pl->pl_fheight) + pl->pl_first;
-			if (nr >= __get_playlist_length())
-				nr = __get_playlist_length() - 1;
-			if (!(event->state & GDK_CONTROL_MASK))
-			{
-				list = get_playlist();
-				while (list)
-				{
-					entry = (PlaylistEntry *) list->data;
-					entry->selected = FALSE;
-					list = list->next;
-				}
-			}
+		int nr, y;
 
-			if (event->state & GDK_SHIFT_MASK && pl->pl_prev_selected != -1)
-			{
-				if (pl->pl_prev_min != -1)
-					playlist_list_select_range(pl->pl_prev_min, pl->pl_prev_max, FALSE);
-				playlist_list_select_range(pl->pl_prev_selected, nr, TRUE);
-				pl->pl_prev_min = pl->pl_prev_selected;
-				pl->pl_prev_max = nr;
-				pl->pl_drag_pos = nr - pl->pl_first;
-			}
-			else
-			{
-				if ((list = g_list_nth(get_playlist(), nr)) != NULL)
-				{
-					entry = list->data;
-					entry->selected = !entry->selected;
-					pl->pl_prev_selected = nr;
-					pl->pl_drag_pos = nr - pl->pl_first;
-					pl->pl_prev_min = -1;
-				}
-			}
-			PL_UNLOCK();
-			if (event->type == GDK_2BUTTON_PRESS)
-			{
-				playlist_set_position(nr);
-				if (!get_input_playing())
-					playlist_play();
-			}
-			pl->pl_dragging = TRUE;
-			playlistwin_update_list();
+		y = event->y - pl->pl_widget.y;
+		nr = (y / pl->pl_fheight) + pl->pl_first;
+		if (nr >= get_playlist_length())
+			nr = get_playlist_length() - 1;
+		if (!(event->state & GDK_CONTROL_MASK))
+			playlist_select_all(FALSE);
+		
+		if (event->state & GDK_SHIFT_MASK && pl->pl_prev_selected != -1)
+		{
+			playlist_select_range(pl->pl_prev_selected, nr, TRUE);
+			pl->pl_prev_min = pl->pl_prev_selected;
+			pl->pl_prev_max = nr;
+			pl->pl_drag_pos = nr - pl->pl_first;
 		}
+		else
+		{
+			if (playlist_select_invert(nr))
+			{
+				pl->pl_prev_selected = nr;
+				pl->pl_drag_pos = nr - pl->pl_first;
+				pl->pl_prev_min = -1;
+			}
+		}
+		if (event->type == GDK_2BUTTON_PRESS)
+		{
+			/*
+			 * Ungrab the pointer to prevent us from
+			 * hanging on to it during the sometimes slow
+			 * playlist_play().
+			 */
+			gdk_pointer_ungrab(GDK_CURRENT_TIME);
+			gdk_flush();
+			playlist_set_position(nr);
+			if (!get_input_playing())
+				playlist_play();
+		}
+		pl->pl_dragging = TRUE;
+		playlistwin_update_list();
 	}
 }
 
+int playlist_list_get_playlist_position(PlayList_List *pl, int x, int y)
+{
+	int iy, length;
+	
+	if (!inside_widget(x, y, pl) || !pl->pl_fheight)
+		return -1;
+
+	if ((length = get_playlist_length()) == 0)
+		return -1;
+	iy = y - pl->pl_widget.y;
+
+	return(MIN((iy / pl->pl_fheight) + pl->pl_first, length - 1));
+}
+
 void playlist_list_motion_cb(GtkWidget * widget, GdkEventMotion * event, PlayList_List * pl)
 {
 	gint nr, y, off, i;
@@ -211,9 +218,9 @@ void playlist_list_motion_cb(GtkWidget * widget, GdkEventMotion * event, PlayLis
 			for (i = 0; i < abs(off); i++)
 			{
 				if (off < 0)
-					playlist_list_move_up();
+					playlist_list_move_up(pl);
 				else
-					playlist_list_move_down();
+					playlist_list_move_down(pl);
 
 			}
 			playlistwin_update_list();
@@ -231,10 +238,39 @@ void playlist_list_button_release_cb(GtkWidget * widget, GdkEventButton * event,
 
 #ifdef HAVE_WCHAR_H
 
+static GdkWChar * find_in_wstr(GdkWChar *haystack, char * needle)
+{
+	/* This will only work if needle is 7bit ASCII characters only */
+	GdkWChar *tmp = haystack;
+	int i = 0;
+
+	if (haystack == NULL)
+		return NULL;
+
+	if (needle == NULL || *needle == '\0')
+		return haystack;
+
+	for (; *tmp != L'\0'; tmp++)
+	{
+		if (*tmp == needle[i])
+		{
+			if (needle[i + 1] == '\0')
+				return (tmp - i);
+			i++;
+		}
+		else if (i > 0)
+		{
+			tmp -= i;
+			i = 0;
+		}
+	}
+	return NULL;
+}
+
+
 void playlist_list_draw_string_wc(PlayList_List *pl, GdkFont *font, gint line, gint width, gchar *text)
 {
-	wchar_t *wtext, *wtmp;
-	GdkWChar *gwtext;
+	GdkWChar *wtext;
 	int len, newlen;
 	/*
 	 * Convert the string to a wide character string to avoid
@@ -245,52 +281,40 @@ void playlist_list_draw_string_wc(PlayList_List *pl, GdkFont *font, gint line, g
 	 * Allocate some extra space, we might extend it by one
 	 * character below
 	 */
-	wtext = g_malloc((strlen(text) + 3)
-			 * sizeof(wchar_t));
-	len = mbstowcs(wtext, text, strlen(text) + 1);
+	wtext = g_malloc((strlen(text) + 3) * sizeof(GdkWChar));
+	len = gdk_mbstowcs(wtext, text, strlen(text) + 1);
 	if (len == -1)
 	{
-		char *tmp;
-		g_message("mbstowcs() failed. (string: %s)", text);
-		for (tmp = text; *tmp; *(tmp++) &= 0x7F);
-		mbstowcs(wtext, text, strlen(text) + 1);
+		/* Conversion failed */
+		for (len = 0; text[len] != '\0'; len++)
+			wtext[len] = text[len];
 	}
-	else
-		wtext[len] = L'\0';
+	wtext[len] = L'\0';
 	if (cfg.convert_underscore)
-		while ((wtmp = wcschr(wtext, L'_')) != NULL)
-			*wtmp = L' ';
-	if (cfg.convert_twenty)
-		while ((wtmp = wcsstr(wtext, L"%20")) != NULL)
+	{
+		int i;
+		for (i = 0; i < len; i++)
+			if (wtext[i] == L'_')
+				wtext[i] = L' ';
+	}
+
+	if (cfg.convert_twenty && len > 2)
+	{
+		GdkWChar *wtmp;
+		while ((wtmp = find_in_wstr(wtext, "%20")) != NULL)
 		{
-			wchar_t *wtmp2 = wtmp + 3;
+			GdkWChar *wtmp2 = wtmp + 3;
 			*(wtmp++) = L' ';
 			while (*wtmp2)
 				*(wtmp++) = *(wtmp2++);
 			*wtmp = L'\0';
+			len -= 2;
 		}
-	len = wcslen(wtext);
-
-	/*
-	 * Convert the wide string to a GdkWChar string that someone,
-	 * for some unknown reason has desided to make uncompatible
-	 * with wchar_t strings.
-	 */
-	if (sizeof(wchar_t) != sizeof(GdkWChar))
-	{
-		int i;
-
-		gwtext = g_malloc((len + 3) * sizeof(GdkWChar));
-		for (i = 0; i <= len; i++)
-			gwtext[i] = wtext[i];
-		g_free(wtext);
 	}
-	else
-		gwtext = (GdkWChar*) wtext;
-	
+		
 	newlen = len + 2;
-	
-	while (gdk_text_width_wc(font, gwtext, len) > width && len > 4)
+
+	while (gdk_text_width_wc(font, wtext, len) > width && len > 4)
 	{
 		/*
 		 * First check if the string gets short enough by
@@ -299,14 +323,17 @@ void playlist_list_draw_string_wc(PlayList_List *pl, GdkFont *font, gint line, g
 		 * remove charaters, one by one.
 		 */
 		len = newlen--;
-		gwtext[len - 3] = L'.';
-		gwtext[len - 2] = L'.';
-		gwtext[len - 1] = L'.';
-		gwtext[len] = L'\0';
+		wtext[len - 3] = L'.';
+		wtext[len - 2] = L'.';
+		wtext[len - 1] = L'.';
+		wtext[len] = L'\0';
 	}
 	
-	gdk_draw_text_wc(pl->pl_widget.parent, font, pl->pl_widget.gc, pl->pl_widget.x, pl->pl_widget.y + line * pl->pl_fheight + font->ascent, gwtext, len);
-	g_free(gwtext);
+	gdk_draw_text_wc(pl->pl_widget.parent, font, pl->pl_widget.gc,
+			 pl->pl_widget.x,
+			 pl->pl_widget.y + line * pl->pl_fheight + font->ascent,
+			 wtext, len);
+	g_free(wtext);
 }
 
 #else /* !HAVE_WCHAR_H */
@@ -345,12 +372,37 @@ void playlist_list_draw_string(PlayList_List *pl, GdkFont *font, gint line, gint
 	gdk_draw_text(pl->pl_widget.parent, font, pl->pl_widget.gc, pl->pl_widget.x, pl->pl_widget.y + line * pl->pl_fheight + font->ascent, text, len);
 }
 
+#if 0
+static GdkPixmap *get_transparency_pixmap()
+{
+	Atom prop, type; 
+	int format;
+	unsigned long length, after;
+	unsigned char *data;
+	static GdkPixmap *retval = NULL;
+	
+	if(retval)
+		return retval;
+
+	prop = XInternAtom(GDK_DISPLAY(), "_XROOTPMAP_ID", True);
+	
+	if(prop==None)
+		return NULL;
+	
+	XGetWindowProperty(GDK_DISPLAY(), GDK_ROOT_WINDOW(), prop, 0L, 1L, False, AnyPropertyType, &type, &format, &length, &after, &data);
+
+	if(type==XA_PIXMAP)
+		retval=gdk_pixmap_foreign_new(*((Pixmap *)data));
+		
+	return retval;
+}
+#endif
+
 void playlist_list_draw(Widget * w)
 {
 	PlayList_List *pl = (PlayList_List *) w;
 	GdkGC *gc;
 	GdkPixmap *obj;
-	GdkFont *font = NULL;
 	gint width, height;
 	gchar *text, *title;
 	gint i, tw, max_first;
@@ -361,15 +413,22 @@ void playlist_list_draw(Widget * w)
 
 	obj = pl->pl_widget.parent;
 
+#if 0
+	if(cfg.playlist_transparent)
+		gdk_draw_pixmap(obj, gc, get_transparency_pixmap(), cfg.playlist_x + pl->pl_widget.x, cfg.playlist_y + pl->pl_widget.y, pl->pl_widget.x, pl->pl_widget.y, width, height);
+/*  	else  */
+/*  	{ */
+#endif
 	gdk_gc_set_foreground(gc, get_skin_color(SKIN_PLEDIT_NORMALBG));
 	gdk_draw_rectangle(obj, gc, TRUE, pl->pl_widget.x, pl->pl_widget.y, width, height);
+/*  	} */
 
-	if ((font = util_font_load(cfg.playlist_font)) != NULL)
+	if (playlist_list_font != NULL)
 	{
 		GList *list;
 		PL_LOCK();
 		list = get_playlist();
-		pl->pl_fheight = font->ascent + font->descent + 1;
+		pl->pl_fheight = playlist_list_font->ascent + playlist_list_font->descent + 1;
 		pl->pl_num_visible = height / pl->pl_fheight;
 
 		max_first = (g_list_length(list) - pl->pl_num_visible);
@@ -406,8 +465,8 @@ void playlist_list_draw(Widget * w)
 				gchar length[20];
 				
 				sprintf(length, "%d:%-2.2d", entry->length / 60000, (entry->length / 1000) % 60);
-				gdk_draw_text(obj, font, gc, pl->pl_widget.x + width - gdk_text_width(font, length, strlen(length)) - 2, pl->pl_widget.y + (i - pl->pl_first) * pl->pl_fheight + font->ascent, length, strlen(length));
-				tw = width - gdk_text_width(font, length, strlen(length)) - 5;
+				gdk_draw_text(obj, playlist_list_font, gc, pl->pl_widget.x + width - gdk_text_width(playlist_list_font, length, strlen(length)) - 2, pl->pl_widget.y + (i - pl->pl_first) * pl->pl_fheight + playlist_list_font->ascent, length, strlen(length));
+				tw = width - gdk_text_width(playlist_list_font, length, strlen(length)) - 5;
 			}
 			else
 				tw = width;
@@ -417,16 +476,15 @@ void playlist_list_draw(Widget * w)
 				text = g_strdup_printf("%s", title);
 
 			if (cfg.use_fontsets)
-				playlist_list_draw_string_wc(pl, font, i - pl->pl_first, tw, text);
+				playlist_list_draw_string_wc(pl, playlist_list_font, i - pl->pl_first, tw, text);
 			else
-				playlist_list_draw_string(pl, font, i - pl->pl_first, tw, text);
+				playlist_list_draw_string(pl, playlist_list_font, i - pl->pl_first, tw, text);
 			g_free(text);
 
 			list = list->next;
 			i++;
 		}
 		PL_UNLOCK();
-		gdk_font_unref(font);
 	}
 	else
 	{
@@ -457,3 +515,11 @@ PlayList_List *create_playlist_list(GList ** wlist, GdkPixmap * parent, GdkGC *
 	add_widget(wlist, pl);
 	return pl;
 }
+
+void playlist_list_set_font(char *font)
+{
+	if (playlist_list_font)
+		gdk_font_unref(playlist_list_font);
+
+	playlist_list_font = util_font_load(font);
+}
